"use strict";
/**
 * @file http
 * @author david wang
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var axios_1 = require("axios");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var enum_1 = require("./enum");
var const_1 = require("./const");
var response_1 = require("./response");
var pubsub_1 = require("./pubsub");
var HttpClient = /** @class */ (function () {
    function HttpClient(config) {
        if (config === void 0) { config = {}; }
        /** @private 公共头部 */
        this._commonHeader = {};
        /** @private 全局配置项 */
        this._config = {};
        this._pubsub = new pubsub_1.PubSub();
        this._config = __assign({}, const_1.HTTP_DEFAULT_CONFIG, config);
    }
    /**
     * @private 判断是否是合法的请求方法
     * @param { string } method 方法名
     * @return { boolean } true or false
     */
    HttpClient.prototype._isValidMethod = function (method) {
        return [
            'delete',
            'get',
            'head',
            'options',
            'patch',
            'post',
            'put',
            'jsonp'
        ].includes(method);
    };
    /**
     * @private 参数转换
     * @param { object } params 要转换的参数
     * @return { string } 转换的值
     */
    HttpClient.prototype._convertParam = function (params) {
        if (!params)
            return '';
        var queryKeys = Object.keys(params);
        var array = [];
        for (var _i = 0, queryKeys_1 = queryKeys; _i < queryKeys_1.length; _i++) {
            var key = queryKeys_1[_i];
            if (typeof params[key] !== 'undefined') {
                array.push(encodeURIComponent(key) + "=" + encodeURIComponent(params[key]));
            }
        }
        return array.join('&');
    };
    /**
     * @private 获取请求路径
     * @param { string } url 路径
     * @param { string } params 参数
     * @return { string } url
     */
    HttpClient.prototype._getUrl = function (url, params) {
        if (params === void 0) { params = ''; }
        if (params)
            url += /\?$/.test(url) ? "" + params : "?" + params;
        return url;
    };
    /**
     * @private 获取header
     * @param { object } header header
     * @return { object } header
     */
    HttpClient.prototype._getHeader = function (header) {
        if (header === void 0) { header = {}; }
        return __assign({}, this._commonHeader, header);
    };
    /**
     * @private 发送请求
     * @param { string } method 请求方法
     * @param { string } url 请求地址
     * @param { HttpRequestOptionType } options 配置
     * @return { object } < Observable<any> >
     */
    HttpClient.prototype._request = function (method, url, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        return new rxjs_1.Observable(function (observer) {
            if (!url)
                throw new Error('the url is required!');
            //校验是否是合法的方法
            if (!_this._isValidMethod(method))
                throw new Error("method " + method + " is not valid!");
            //获取参数
            var params;
            if (!!options.params)
                params = _this._convertParam(options.params);
            //获取请求路径
            url = _this._getUrl(url, params);
            //获取header
            var header = _this._getHeader(options.header);
            //cancel Token
            var cancelToken = axios_1.default.CancelToken;
            var cancelSource = cancelToken.source();
            //发送请求
            axios_1.default.request({
                method: method,
                url: url,
                baseURL: options.baseUrl,
                data: options.body,
                headers: header,
                withCredentials: !!options.withCredentials,
                responseType: options.responseType || 'json',
                onUploadProgress: function (progressEvent) {
                    observer.next(new response_1.HttpUploadProgressEvent(progressEvent));
                },
                onDownloadProgress: function (progressEvent) {
                    observer.next(new response_1.HttpDownloadProgressEvent(progressEvent));
                },
                cancelToken: cancelSource.token,
                validateStatus: function (status) {
                    var validateStatus = _this._config.validateStatus;
                    if (typeof validateStatus === 'function')
                        return validateStatus(status);
                    console.warn('the option validateStatus must be function that return boolean!');
                    return false;
                }
            })
                .then(function (response) {
                observer.next(response);
                observer.complete();
            })
                .catch(function (error) {
                observer.error(new response_1.HttpResponseError(error.response || error.request || error, options));
            });
            // 返回一个取消订阅的函数
            return function () {
                //取消请求
                cancelSource.cancel();
            };
        });
    };
    /**
     * 处理json返回数据
     * @param { object } json 返回的json值
     * @param { object } response response
     * @param { object } options 当前请求的配置项
     * @return { any } 返回object|httpResponseError
     */
    HttpClient.prototype._handleJsonResponse = function (json, response, options) {
        var responseKey = options.responseResultKey;
        var successCode = options.successCode;
        var keyCode = responseKey.code;
        //判断是否是成功的
        if (!Array.isArray(successCode))
            successCode = [successCode];
        if (json &&
            typeof json[keyCode] !== 'undefined' &&
            successCode.indexOf(json[keyCode]) !== -1) {
            // --todo http 缓存
            //返回数据内容
            return json[responseKey.data];
        }
        else {
            //抛出异常
            return new response_1.HttpResponseError(response, options);
        }
    };
    /**
     * 错误返回处理
     * @param { any } err 错误对象
     * @param { object } options 配置项
     * @return { void }
     */
    HttpClient.prototype._handleResponseError = function (err, options) {
        //错误码
        var status = err.status;
        var errorType;
        //处理一些常见的 status
        if (status >= 500)
            errorType = enum_1.HttpErrorCodeType.Server;
        if (status === 0)
            errorType = enum_1.HttpErrorCodeType.Offline;
        if (status === 401)
            errorType = enum_1.HttpErrorCodeType.UnAuth;
        if (status === 404)
            errorType = enum_1.HttpErrorCodeType.NotFound;
        this._pubsub.emit(errorType, err, options);
    };
    /**
     * 设置公共头
     * @param { string } key header key
     * @param { any } value header value
     * @return { void }
     */
    HttpClient.prototype.setCommonHeader = function (key, value) {
        this._commonHeader[key] = value;
    };
    /**
     * 监听
     */
    HttpClient.prototype.listen = function (handle) {
        var _loop_1 = function (key) {
            var type = enum_1.HttpErrorCodeType[key];
            var callback = function () {
                return function (err, options) {
                    handle(type, err, options);
                };
            };
            this_1._pubsub.on(type, callback());
        };
        var this_1 = this;
        for (var key in enum_1.HttpErrorCodeType) {
            _loop_1(key);
        }
        return this._pubsub.clear;
    };
    /**
     * @private 发送请求
     * @param { string } method 请求方法
     * @param { string } url 请求地址
     * @param { HttpRequestOptionType } options 配置
     * @return { Object } < Observable<any> >
     */
    HttpClient.prototype.request = function (method, url, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        var config = this._config;
        //拼合配置项
        options = __assign({}, const_1.HTTP_DEFAULT_REQUEST_OPTION, config, options);
        return this._request(method, url, options).pipe(operators_1.filter(function (response) {
            //不要求跟踪事件，那么就过滤它
            if (options.observe !== enum_1.HttpObserveType.Event) {
                return !(response instanceof response_1.HttpDownloadProgressEvent ||
                    response instanceof response_1.HttpUploadProgressEvent);
            }
            return true;
        }), operators_1.map(function (response) {
            if (options.observe === enum_1.HttpObserveType.Body) {
                var data = response.data;
                var result = data;
                //如果返回的是json数据
                if (options.responseType === 'json' &&
                    typeof data === 'object') {
                    result = _this._handleJsonResponse(data, response, options);
                    //如果结果不成功
                    if (result instanceof response_1.HttpResponseError) {
                        throw result;
                    }
                    else {
                        return { result: result, response: response, isSuccess: true };
                    }
                }
                return { result: result, response: response, isSuccess: true };
            }
            return response;
        }), operators_1.catchError(function (error) {
            if (error instanceof response_1.HttpResponseError) {
                _this._handleResponseError(error, options);
            }
            else {
                console.error(error);
            }
            //返回null
            return rxjs_1.of({ result: null, response: error, isSuccess: false });
        }));
    };
    /**
     * post请求
     * @param body
     * @param options
     */
    HttpClient.prototype.post = function (url, body, options) {
        if (body === void 0) { body = {}; }
        if (options === void 0) { options = {}; }
        return this.request("post", url, __assign({}, options, { body: body }));
    };
    /**
     * get
     * @param params
     * @package options
     */
    HttpClient.prototype.get = function (url, params, options) {
        if (params === void 0) { params = {}; }
        if (options === void 0) { options = {}; }
        return this.request("get", url, __assign({}, options, { params: params }));
    };
    /**
     * put
     * @param params
     * @package options
     */
    HttpClient.prototype.put = function (url, params, options) {
        if (params === void 0) { params = {}; }
        if (options === void 0) { options = {}; }
        return this.request("put", url, __assign({}, options, { params: params }));
    };
    /**
     * delete
     * @param params
     * @package options
     */
    HttpClient.prototype.delete = function (url, params, options) {
        if (params === void 0) { params = {}; }
        if (options === void 0) { options = {}; }
        return this.request("delete", url, __assign({}, options, { params: params }));
    };
    return HttpClient;
}());
exports.default = {
    createClient: function (config) {
        return new HttpClient(config);
    }
};
